@echo off
setlocal ENABLEEXTENSIONS

set PYTHON=python

@echo on
%PYTHON% test_0.py
%PYTHON% test_1.py
%PYTHON% test_2a.py
%PYTHON% test_2b.py
%PYTHON% test_3.py
%PYTHON% test_4.py
%PYTHON% test_5.py
