cdef Any
cdef Union
cdef Literal
cdef Optional
cdef NoReturn

cdef Callable
cdef Hashable
cdef Iterable
cdef Iterator
cdef Sequence
cdef Mapping

cdef Tuple
cdef List
cdef Dict

cdef Buffer

cdef BufSpec
cdef BufSpecB
cdef BufSpecV
cdef BufSpecW
cdef TargetSpec
